#include "inc.h"
#include "arithc.h"

/* consts from Michael Schindler */

#define CODE_BITS               31
#define SHIFT_BITS              (CODE_BITS - 8)
#define CODE_BYTES              ((CODE_BITS+7)/8)

#define PRECISION_BITS  9               // coding is done to this accuracy (in terms of range>>PRECISION_BITS)

#define MinRange                ((ulong)1<<SHIFT_BITS)
#define One                     ((ulong)1<<CODE_BITS)
#define CumProbMax      (MinRange>>PRECISION_BITS)
#define CODE_MASK               (One - 1)

#define EXTRA_BITS              ((CODE_BITS-1) % 8 + 1)         // == 7 == CODE_BITS - (8*(CODE_BYTES-1))
#define TAIL_EXTRA_BITS (8 - EXTRA_BITS)                        // == 1

        // extra_bits are the bits in "code" that don't quite fit in bytes

//--------------------------------------------------------------
//--- range normalizers for encoder & decoder; also does the IO

static void enc_renorm( Arith* arith, ulong code, ulong range ) {
    assert( range <= One );

    while (range <= MinRange) {

        ulong byte = (code >> SHIFT_BITS);
        
        if (byte == 0xFF) {

            /** the waiting queue is incremented like :
             *               (arith->queue), 0xFF, 0xFF, ... ,0xFF, code
             ***/

            arith->overflow_bytes++;

        } else {

            ulong carry = code>>CODE_BITS;        

            /* carry == 1 or 0 : is the top bit on ?
             *       carry = byte>>8
             * if ( carry )  send nextbyte+1
             *                               MinRange queue with zeros
             * else                  send nextbyte
             *                               MinRange queue with ones
             **/                                     
                
            *(arith->outPtr)++ = (ubyte)(arith->queue + carry); // propagate the carry.
            // send the queue
            if (arith->overflow_bytes) {
                *(arith->outPtr)++ = (ubyte)(0xFF + carry);
                while ( --(arith->overflow_bytes) ) *(arith->outPtr)++ = (ubyte)(0xFF + carry);
            }
            arith->queue = byte;
        }

        code = (code<<8) & CODE_MASK;
        range <<= 8;
    }

    assert( range <= One );

    arith->code  = code;
    arith->range = range;
}

/* */
static void dec_renorm( Arith* arith, ulong* pcode, ulong* prange ) {

    ulong range = arith->range;
    ulong code  = arith->code;

    assert( range <= One );

    while (range <= MinRange) {
        range <<= 8;
        code = (code<<8) + (((arith->queue)<<EXTRA_BITS)&0xFF);   // use the top bit in the queue
        arith->queue = *(arith->outPtr)++;
        code += (arith->queue) >> (TAIL_EXTRA_BITS);
    }
        
    assert( range <= One );

    *prange = range;
    *pcode = code;
}

/* 2004-04-25-CrT: Killed to quiet gcc warning: */
/* #pragma warning( disable : 4244 )    // loss of data in conversion */

//#define PARANOID_ENCODEDONE   // @@ ?

//#define FAST_ENCODE
        /** actually makes very little difference in speed or compression
        *       presumably because the Pentium's branch prediction is doing
        *       a good job
        **/

/* */
Arith* arith_init( void ) {
    Arith* arith;

    if ((arith = new(Arith)) == NULL)   return NULL;

    arith->probMax     = CumProbMax;
    arith->probMaxSafe = CumProbMax - 512;

    return arith;
}

/* */
void arith_destroy( Arith* arith ) {   if (arith) free(arith);   }

/* */
void arith_encodeInitNoStuff( Arith* arith, ubyte* outBuf ) {
    arith->outBuf = outBuf;
    arith_encodeReInit( arith, outBuf - 1 );
}

/* */
void arith_encodeInit( Arith* arith, ubyte* outBuf ) {
    arith->outBuf = outBuf;
    arith_encodeReInit( arith, outBuf );
}

/* */
void arith_encodeReInit( Arith* arith, ubyte* outBuf ) {

    arith->code  = 0;
    arith->range = One;
    arith->overflow_bytes = 0;
    arith->queue = 0; // this is a waste of a byte

    /* arith->outBuf unchanged: */
    arith->outPtr = outBuf;
}

/* */
void arith_encode( Arith* arith, ulong symlow, ulong symhigh, ulong symtot ) {
    ulong code  = arith->code;
    ulong range = arith->range;

    assert( symlow < symhigh && symhigh <= symtot );
    assert( symtot <= CumProbMax );

    /** we want to do :
     *       lowincr = (range * symlow) / symtot
     *       but range & symtot can both be large , so this would overflow the register      
     *       thus we instead do:
     *
     ***/

#ifdef FAST_ENCODE
    {   ulong r;

        r = range / symtot;
        
        code += r * symlow;
        range = r * (symhigh - symlow);

    }
#else
    { 
        ulong lowincr,r;

        r = range / symtot;

        lowincr = r * symlow;
        code += lowincr;
        if (symhigh == symtot)   range -= lowincr;
        else                     range = r * (symhigh - symlow);

    }
#endif

    enc_renorm( arith, code, range );
}

/* */
int arith_encodeDone( Arith* arith ) {
    uint rangeMask;
    uint rangeMSB;

    /* Set code to the maximum that won't change how it decodes: */

    arith->code += arith->range - 1;

    /* First send the queue: */
    if (arith->code & One) {
        *(arith->outPtr)++ = (ubyte)(arith->queue + 1)&0xFF;
        while ( arith->overflow_bytes-- ) *(arith->outPtr)++ = 0;
    } else {
        *(arith->outPtr)++ = arith->queue;
        while ( arith->overflow_bytes-- ) *(arith->outPtr)++ = 0xFF;
    }

    /*****

    The minimal way to flush is to do :

                code += (range - 1);

                clear code below MSB of range

    eg. if range is 67 we do :
                
                code += 66;
                code &= ~63;

        then we just send code bytes until the remainder is zero.

        (this assumes that when the decoder reads past EOF, it reads zeros!)

        -----

        we almost always write just 1 byte
        (in fact, I think we might *always* write 1 byte)

        ******/

    if (arith->range >= (1UL<<31)) {
        rangeMSB = 1UL<<31;
    } else {
        for (rangeMSB = 1;   rangeMSB <= arith->range;   rangeMSB <<= 1) ;
        rangeMSB >>= 1;
        assert( arith->range >= rangeMSB && arith->range < rangeMSB*2 );
    }

        // Clear code under rangeMask :
    rangeMask = 0;
    rangeMSB >>= 1;
    while (rangeMSB) {
        rangeMask |= rangeMSB;
        rangeMSB >>= 1;
    }

    assert( rangeMask < arith->range );

    arith->code &= (~ rangeMask);
    arith->code &= CODE_MASK;

    while (arith->code) {
        *(arith->outPtr)++ = (arith->code >> SHIFT_BITS) & 0xFF;
        arith->code <<= 8;
        arith->code &= CODE_MASK;
    }

    arith->outPtr[0] = 0;
    arith->outPtr[1] = 0;
    arith->outPtr[2] = 0;
    arith->outPtr[3] = 0;
    arith->outPtr[4] = 0;
    arith->outPtr[5] = 0;

    return arith->outPtr - arith->outBuf;
}

/* */
void arith_decodeInitNoStuff( Arith* arith, ubyte* outBuf ) {
    arith->outBuf = arith->outPtr = outBuf;
    arith_decodeReInit(arith,outBuf-1);
}

/* */
void arith_decodeInit( Arith* arith, ubyte* outBuf )   {
    arith->outBuf = arith->outPtr = outBuf;
    arith_decodeReInit(arith,outBuf);
}

/* */
void arith_decodeReInit( Arith* arith, ubyte* outBuf ) {
    ulong byte;

    arith->outPtr = outBuf + 1;

    /** the code needs to be kept filled with 31 bits ;
     *       this means we cannot just read in 4 bytes.  We must read in 3,
     *       then the 7 bits of another (EXTRA_BITS == 7) , and save that last
     *       bit in the queue 
     **/

    byte = arith->queue = *(arith->outPtr)++;
    arith->code = byte >> (TAIL_EXTRA_BITS);
    arith->range = 1 << EXTRA_BITS;
}

/** Save a repeated divide computation between arith_get and arith_decode **/
static ulong dec_range_over_symtot;

/* */
ulong arith_get( Arith* arith, ulong symtot ) {
    ulong ret;
    ulong range;
    ulong code;

    assert( symtot <= CumProbMax );

    /* code, range = */ dec_renorm( arith, &code, &range );

    dec_range_over_symtot = range / symtot;
    ret = code / dec_range_over_symtot;

    arith->range = range;
    arith->code  = code;
                
    //assert( ret < symtot );  // is this safe? ; apparently not!
    ret = ( ret >= symtot ? symtot-1 : ret );

    return ret;
}

/* */
void arith_decode( Arith* arith, ulong symlow, ulong symhigh, ulong symtot ) {

    assert( symlow < symhigh && symhigh <= symtot );
    assert( symtot <= CumProbMax );

#ifdef FAST_ENCODE

    arith->code -= dec_range_over_symtot * symlow;
    arith->range = dec_range_over_symtot * (symhigh - symlow);

#else
    {   ulong lowincr;

        lowincr = dec_range_over_symtot * symlow;
        arith->code -= lowincr;
        if ( symhigh == symtot )        arith->range -= lowincr;
        else                            arith->range = dec_range_over_symtot * (symhigh - symlow);
    }
#endif

}

/* */
void arith_decodeDone( Arith* arith ) { }

/* */
void arith_encBit( Arith* arith,   ulong mid,   ulong tot,  bool bit ) {
    ulong code  = arith->code;
    ulong range = arith->range;

    {    ulong r = (range / tot) * mid;

        if (bit) {
            code += r;
            range -= r;
        } else {
            range = r;
        }
    }

    enc_renorm( arith, code, range );
}

/* */
bool arith_decBit( Arith* arith, ulong mid, ulong tot ) {
    bool bit;
    ulong range, code, r;

    dec_renorm( arith, &code, &range );

    /**
     *       
     *               r = (range / tot)
     *
     *       if ( (code / r) >= mid ) bit = 1; else bit = 0;
     *
     *       we eliminate one divide. this is the savings of the binary coder
     *
     **/

    r = (range / tot) * mid;

    if (code >= r) {
        bit = 1;
        code -= r;
        range -= r;
    } else {
        bit = 0;
        range = r;
    }

    arith->range = range;
    arith->code  = code;

    return bit;
}

