
#include "contexttrie.h"
#include "config.h"

/* */
struct ContextTrie {
    Context*  order0;
    Context** order1;
    uint      numNodes;
    uint      orders;
    uint*     alphabets; /* Make this a struct for easier copying? */

    Node  LRU;
    uint      NumLRUContexts;
    uint      MaxLRUContexts;
};

/* */
ContextTrie * contextTrie_create(
    uint* alphabets,
    uint  orders,
    uint  ContextMegabytes
) {
    ContextTrie* t = new( ContextTrie );  assert(t);

    t->orders = orders;
    t->alphabets = malloc( orders * sizeof( uint ) );

    memcpy( t->alphabets, alphabets, orders*sizeof(uint) );

    t->order0 = context_create( NULL, 0 );

    t->order1 = malloc( sizeof(void*) * alphabets[0] );
    {   uint i;
        for (i = 0;   i < alphabets[0];   i++) {
            t->order1[i] = context_create( t->order0, i );
        }
    }

    node_null( &t->LRU );

    t->NumLRUContexts = 0;
    t->MaxLRUContexts = (ContextMegabytes << 20) / sizeof( Context );

    return t;
}

/* */
void contextTrie_destroy( ContextTrie* t ) {
    free( t->alphabets );
    destroy( t );
}

/* */
static void delete_context( ContextTrie* t, Context* c ) {

    assert( ! c->parent || c->parent->order == (c->order - 1) );
    assert( c->order == PZIP_ORDER || ! c->child  || c->child->order  == (c->order + 1) );
    assert( ((Context *)node_next(c))->order == c->order );

    if (c->child) {
        if (c->order == PZIP_ORDER) {
            // c->child is a Det !
            // Det_DeleteNodes(
            c->child = NULL; // just let the det LRU clean it up later
        } else {
            do {
                assert(c->child->order == (c->order + 1) );
                assert(c->child->parent == c);
                delete_context(t,c->child);
            } while(c->child);
        }
    }

    if (c->parent)    {
        assert( c->parent->order == (c->order - 1) );
        if (c->parent->child == c) {
            // Parent points to me. Try to get a sibling:
            c->parent->child = node_next(c);
            if ( c->parent->child == c ) {
                /* No siblings, so nullify: */
                c->parent->child = NULL;
            }
        }
    }

    assert( ! c->parent || c->parent->order == (c->order - 1) );
    assert( ! c->child );
    assert( ((Context*)node_next(c))->order == c->order );

    context_destroy(c); /* Does node_cut(c) & node_cuts the LRU for us */

    -- t->NumLRUContexts;
}

/* */
static void lru_new_context( ContextTrie* t, Context* c ) {
    node_addHead( &t->LRU, &c->LRU );

    ++ t->NumLRUContexts;

    if (t->NumLRUContexts >= t->MaxLRUContexts) {
        Node * pLast;
        
        /* Delete one at tail: */

        pLast = node_cutTail( &t->LRU );
        assert(pLast);
        c = (Context *)( (ulong)pLast - (ulong)(&((Context *)0)->LRU) );

        assert( ! c->parent || c->parent->order == (c->order - 1) );
        assert( c->order == PZIP_ORDER || ! c->child  || c->child->order  == (c->order + 1) );
        assert( ((Context *)node_next(c))->order == c->order );
        assert( c->child != c->parent );

        delete_context(t,c);
    }
}

/* */
static void lru_old_context( ContextTrie* t, Context* c ) {
    node_cut( &c->LRU );
    node_addHead( &t->LRU, &c->LRU );
}

/* */
static Context* find_or_create( ContextTrie* t, Context* c, ulong index ) {
    Context * next;
    next = context_find(c,index);
    if (next) {
        c = next;
        lru_old_context( t, c );
    } else {
        c = context_create(c,index);
        lru_new_context(t,c);
    }
    return c;
}

/* */
void contextTrie_getNodes( ContextTrie* t, ulong* pindex, Context** pinto ) {
    Context * c;
    uint i;

    *pinto++ = c = t->order0;
    *pinto++ = c = t->order1[*pindex++];
    for (i = 2;   i <= t->orders;  i++) {
        c = find_or_create(t,c,*pindex++);

        assert( ! c->parent || c->parent->order == (c->order - 1) );
        assert( c->order == PZIP_ORDER || ! c->child  || c->child->order  == (c->order + 1) );
        assert( ((Context *)node_next(c))->order == c->order );

        *pinto++ = c;
    }
}

