
#include "det.h"
#include "escape.h"
#include "inc.h"
#include "pool.h"
#include "intmath.h"
#include "list.h"
#include "config.h"

/*******

@@ LRU the contexts!

-------------

 with the matchPtr method :

 trans                :    93695 ->    14209 = 1.213 bpb

 with the fast followNode method

 currently -> 14213 , 1.214 bpb

*********/

#define Det_MaxMatchLen      (1024)
#define Det_MaxNodeWalk      (100)

#define NODE_ARRAY_SIZE           (1<<18) // a 256k window

#define DO_ADDNODE_ON_SUCCESS

#define HASH_MASK       (0xFFFF)
#define HASH_SIZE       (HASH_MASK + 1 )
#define HASH(x)         ((((x)>>11) ^ (x) )&HASH_MASK)

typedef struct DetNode       DetNode;
typedef struct DetContext    DetContext;

struct DetNode {
    Node   node;              /* Must be at head. */
    uword  detMinLen;
    uword  sym;
    ubyte* backPtrMinus13;
};

struct DetContext {
    Node nodes;
    uint match;
    uint escape;
};

struct Det {
    Pool*    detContextPool;
    Escape* zesc;

    DetNode  nodeArray[ NODE_ARRAY_SIZE +1 ];
    uint     nodeArrayI;

    /* Stuff saved from Enc/Dec to Update: */

    Context*    gotContext;
    DetContext* gotDContext;
    DetNode*    gotNode;
    DetNode*    nextNode;
    uint        gotML;
    uint        longestML;
};


Det* det_create( uint UseMegs ) {
    int i;
    Det* det = new( Det );
    det->detContextPool = pool_create( sizeof( DetContext ), 1024, 256 );
    det->zesc = escape_create();

    det->nodeArrayI = 0;

    for (i = 0;   i < NODE_ARRAY_SIZE;   i++) {
        node_null( &det->nodeArray[i] );
    }

    return det;
}

void det_destroy( Det* det ) {
    assert( det );
    pool_destroy( det->detContextPool );
    escape_destroy( det->zesc );
    destroy( det );
}

static DetNode* addNodeToContext( Det* det, Context* c, int sym, ubyte* backPtr, uint minLen ) {
    DetNode* node;
    DetContext* dc;

    if (c->child) {
        dc = (DetContext *)(c->child);
    } else {
        dc = pool_getHunk( det->detContextPool, TRUE );
        dc->escape = dc->match = 1;
        node_null( &dc->nodes );
        c->child = (void *)dc;
    }

    node = &det->nodeArray[det->nodeArrayI++];
    if (det->nodeArrayI == NODE_ARRAY_SIZE) {
        det->nodeArrayI = 0;
    }
    node_cut(node);

    if (minLen < DET_MIN_ORDER) {
        minLen = DET_MIN_ORDER;
    }
    node->sym = sym;
    node->detMinLen = minLen;
    node->backPtrMinus13 = backPtr - 13;

    node_null( node );
    node_addHead( &dc->nodes, node );

    return node;
}

void det_update( Det* det, ubyte* backPtr, ubyte* backBase, int sym ) {
    DetNode* node;

    det->nextNode = NULL;

    /* Remembered from the last encode/decode call: */
    node = det->gotNode;

    assert( det->gotContext );

    if (node) {
        assert( det->gotDContext );

        if (node->sym == sym) {
            uint i;

            ++ det->gotDContext->match;

            i = (ulong)node - (ulong)det->nodeArray;

            if (i == (NODE_ARRAY_SIZE-1) * sizeof(DetNode))   det->nextNode = det->nodeArray;
            else                                              det->nextNode = node + 1;

            /* nextNode may not be filled out yet until we do addNode below: */

        } else {

            ++ det->gotDContext->escape;

            assert( det->gotML >= node->detMinLen );

            node->detMinLen = det->gotML + DET_MIN_LEN_INC;
        }
    }
        
    addNodeToContext( det, det->gotContext, sym, backPtr, det->longestML +1 );
}

static int matchBack( ubyte* p1, ubyte* p2,ubyte* backBase ) {
    int len = 0;
    int maxLen = min( (p1 - backBase), (p2 - backBase) );
    maxLen = min( maxLen, Det_MaxMatchLen );
    while (*p1-- == *p2--) {
        len++;
        if (len >= maxLen)   break;
    }
    return len;
}

static DetNode* getNode( Det* det, DetContext* dc, ubyte* backPtr, ubyte* backBase) {
    DetNode *bestNode;
    uint     len;
    uint     bestLen;
    uint     longestLen;
    ubyte*   bm13;
    DetNode* node;
    Node*    list;
    uint     walk;

    if (!dc ) {
        /* @@ cbloom bug fix 5-14-02 */
        det->gotDContext = NULL;
        det->gotNode     = NULL;
        det->longestML   = 0;
        det->gotML       = 0;
        return NULL;
    }

    if ((uint)(backPtr - backBase) < DET_MIN_ORDER)   return NULL;

    bm13     = backPtr - 13;
    bestNode = NULL;
    bestLen  = longestLen = 0;

    list = &dc->nodes;
    walk = 0;
    for (node = node_next(list);   node != (DetNode*)list;   node = node_next(node)) {
        len = matchBack( bm13, node->backPtrMinus13, backBase ) + 12;
        longestLen = max( longestLen, len );
        if (len >= node->detMinLen && len > bestLen) {
            bestLen  = len;
            bestNode = node;
        }
        if (++walk == Det_MaxNodeWalk)   break;
    }
    det->gotDContext = dc;
    det->gotNode     = bestNode;
    det->longestML   = longestLen;
    det->gotML       = bestLen;

    assert( bestLen >= DET_MIN_ORDER || ! bestNode );

    return bestNode;
}

static void do_match( Det* det, ubyte* backPtr, ubyte* backBase, Context* context ) {
    det->gotContext = context;
    if (!det->nextNode) {

        det->gotDContext = NULL;
        det->gotNode = NULL;
        getNode( det, (DetContext*)(context->child), backPtr, backBase );

    } else {

        det->gotDContext = (DetContext*)context->child;
        if (!det->gotDContext) {

            getNode( det, (DetContext*) (context->child), backPtr, backBase );

        } else {

            det->gotNode = det->nextNode;
            det->gotML ++;
            det->longestML = max( det->longestML, det->gotML );
                        
            if (det->gotML >= 64) {
                /* Force it to accept this match: */
                det->gotNode->detMinLen = min( det->gotNode->detMinLen, det->gotML );
            } else {
                if (det->gotML < det->gotNode->detMinLen) {
                    getNode( det, (DetContext*)context->child, backPtr, backBase );
                }
            }
        }
    }
}

bool det_encode( Det* det, Arith* arith, ubyte* backPtr, ubyte* backBase, int sym, Exclude* exc, Context* context, bool* useFull) {
    bool match;
    int  count;
    int  pred;

    assert( *useFull == TRUE );

    do_match( det, backPtr, backBase, context );

    if (!det->gotNode)   return FALSE;

    count = det->gotDContext->match;
    pred =  det->gotNode->sym;
    if (det->gotML >= 64)   count = 99999;

    assert( exclude_isEmpty(exc) );

    match = (sym == pred);

    *useFull = FALSE; // @@ do this?

    // @@ use upex or full for numsyms ?
    escape_encode( det->zesc,arith, getulong(backPtr-4), 1, count, 8, context->upex.numSyms, !match );

    exclude_set( exc, pred );

    return match;
}

bool det_decode( Det* det, Arith* arith, ubyte* backPtr, ubyte* backBase,int* psym, Exclude* exc, Context* context, bool* useFull ) {
    bool match;
    int  count;
    int  pred;

    assert( *useFull == TRUE );

    do_match( det, backPtr, backBase, context );

    if (!det->gotNode)   return FALSE;

    count = det->gotDContext->match;
    pred =  det->gotNode->sym;
    if (det->gotML >= 64) count = 99999;

    assert( exclude_isEmpty(exc) );

    *useFull = FALSE; // @@ do this?

    match = ! escape_decode( det->zesc, arith, getulong( backPtr -4 ), 1, count, 8, context->upex.numSyms );

    exclude_set( exc, pred );

    *psym = pred;

    return match;
}
