
#define _GNU_SOURCE
#include <string.h>
#include <ctype.h>
#include "pzip.h"
#include "config.h"
#include "version.h"
#include "loe.h"
#include "fileutil.h"
#include "inc.h"
#include "crc32.h"


#define PREAMBLE	(1024)

static const ulong PZIP_MAGIC = 0x70707A32; /* "PPZ2" */
 
static int whichByteDiff( ubyte* buf1, ubyte* buf2 ) {
    int b = 0;
    while (buf1[b] == buf2[b]) {
        b++;
    }
    return b;
}

int verbose = FALSE;

int main(  int argc,   char* argv[] ) {
    char*  inName  = NULL;
    char*  outName = NULL;
    ubyte* rawBuf;
    ubyte* compBuf;
    ubyte* decBuf;
    int    compLen;
    int    rawLen;
    bool   encOnly = FALSE;
    bool   encoding= TRUE;
    FILE*  inFP    = NULL;
    FILE*  outFP   = NULL;
    ulong  rawCRC  = 0;

    if (argc < 2) {
	fprintf(stderr, "pzip version %.2f\n", VERSION );
	fprintf(stderr, "Usage : pzip [options] <in> [out]\n" );
	fprintf(stderr, "options :\n" );
	fprintf(stderr, " -e  : encode only [vs also decode and compare]\n");
	fprintf(stderr, " -l# : set loe type to # [MPS]\n");
	fprintf(stderr, " -v  : verbose output during run\n");
	exit(1);
    }

    ++argv;
    --argc;


    /* Process options: */
    while (argc > 0) {
        char*  str = *argv++;
        argc--;

        if (*str == '-') {
            str++;

            switch (*str++) {

            case 'l':
                Pzip_LOEType = atoi(str);
                if (Pzip_LOEType < LOETYPE_NONE  ) Pzip_LOEType = LOETYPE_NONE;
                if (Pzip_LOEType >= LOETYPE_COUNT) Pzip_LOEType = LOETYPE_COUNT-1;
                break;

            case 'e':
                encOnly = TRUE;
                break;

            case 'v':
                ++verbose;
                break;

            default:
                fprintf(stderr,"unknown option : %c : skipped\n",str[-1]);
                break;
            }

        } else {

            if        (!inName)  { inName  = str;
            } else if (!outName) { outName = str;
            } else {
                fprintf(stderr,"extraneous parameter : %s : ignored\n",str);
            }
        }
    }


    inFP = fopen( inName, "r" );
    if (!inFP)   errexit("couldn't read input file");

    rawLen = flength(inFP);

    if (outName) {
        outFP = fopen( outName, "w" );
        if (!outFP)   errexit("couldn't open output file");
    }

    encoding = TRUE;

    if (outFP) {

        /* Is inFP compressed? */
        ulong tag = fgetul( inFP );
        if (tag == PZIP_MAGIC) {
            /* It is packed: */
            rawLen = fgetul( inFP );
            rawCRC = fgetul( inFP );
            encoding = FALSE;
        } else {
            /* Not packed: */
            fseek( inFP, 0, SEEK_SET );
            fputul( PZIP_MAGIC, outFP );
            fputul( rawLen, outFP );
        }
    }

    rawBuf = malloc( rawLen + 1024 + PREAMBLE );
    assert( rawBuf );

    memset( rawBuf, ' ', PREAMBLE );
    rawBuf += PREAMBLE;

    if (encoding) {
        fread( rawBuf, 1, rawLen, inFP );
        fclose(inFP);
        inFP = NULL;

        rawCRC = crcbuf( rawBuf, rawLen );

        if (outFP) { fputul(rawCRC,outFP); }
    }

    decBuf = malloc( rawLen + 1024 + PREAMBLE );
    assert( decBuf );

    memset( decBuf, ' ', PREAMBLE );
    decBuf += PREAMBLE;

    compBuf = malloc( rawLen*2 + 65536 + PREAMBLE );
    assert( compBuf );

    memset( compBuf, 0, PREAMBLE );
    compBuf += PREAMBLE;

    if (encoding) {
        compLen = pzip_encodeArray( rawBuf, rawLen, compBuf );
        if (verbose) {
            fprintf(stderr,
                "%-20s : %8d -> %8d = %1.3f bpc\n",
                basename(inName), rawLen, compLen, compLen * 8.0 / (double) rawLen
            );
        }
    } else {
        /* 12 bytes of header: */
        compLen = flength(inFP) - 12;
        fseek( inFP, 12, SEEK_SET );
        fread( compBuf, 1, compLen, inFP );
        fclose(inFP);
        inFP = NULL; 
    }

    if (!encOnly) {
        ulong decCRC;

        if (!pzip_decodeArray(decBuf,rawLen,compBuf) ) {
            errexit("decode failed");	
        }	
        decCRC = crcbuf(decBuf,rawLen);

        if (decCRC != rawCRC)	{
            fprintf(stderr, "***** decoded crc failure : %08lX vs %08lX\n",rawCRC,decCRC);
        }
    }

    /* Compare rawBuf and decBuf if we have them both: */
    if (encoding && !encOnly) {
        if (memcmp( decBuf, rawBuf, rawLen )) {
            fprintf(stderr, "***** decode failed! %d th bytes differ\n",whichByteDiff(decBuf,rawBuf));
        }
    }

    if (outFP) {
        if (encoding) fwrite( compBuf, 1, compLen, outFP );
        else          fwrite( decBuf,  1, rawLen,  outFP );

        fclose( outFP );
        outFP = NULL;
    }

    return 0;
}
