#include "inc.h"
#include "pool.h"

typedef struct Block {
    struct Block* next;
    char*         base;
    char*         ptr;
    long          length;
    long          free;
} Block;

struct Pool {
    long   hunkLength;

    Block* curBlock;
    Block* block;
    void** freedHunks;

    long   autoExtendNumItems;
    long   numFreedHunks;
    long   maxNumFreedHunks;
    long   numItemsActive;
};

/*
 *  Pool
 *    system for fast Resetting & Freeing of nodes; for use in tree structures
 *     requires no traversing for freeing
 *    does auto-extending of memory space in case you use more space than
 *     expected or if you don't know how much you will need
 *
 */

static bool extend( Pool* pool, long numHunks ) {
    Block* block = pool->block;

    while (block) {
        if (block->free > pool->hunkLength) {
            pool->curBlock = block;
            return 1;
        }
        block = block->next;
    }

    if (!(block = malloc( sizeof( Block ) ))) {
        return 0;
    }

    block->length = numHunks * pool->hunkLength;
    block->free   = block->length;

    if (!(block->base = malloc( block->length ))) {
        free( block );
        return 0;
    }

    block->ptr  = block->base;
    block->next = pool->block;
    pool->block = block;

    pool->curBlock = block;

    return 1;
}

/*
 * hunkLength is forced to be a multiple of 4
 */
Pool* pool_create( long hunkLength, long NumHunks, long autoExtendNumItems ) {
    Pool* pool;

    if ( (pool = malloc(sizeof(Pool))) == NULL ) {
        return NULL;
    }

    pool->hunkLength  = paddedSize(hunkLength);
    pool->curBlock = NULL;
    pool->block    = NULL;

    pool->numFreedHunks      = 0;
    pool->maxNumFreedHunks   = 16;
    pool->autoExtendNumItems = autoExtendNumItems;
    pool->numItemsActive     = 0;

    if (!(pool->freedHunks = malloc( pool->maxNumFreedHunks * sizeof(void*) ))) {
        free( pool );
        return NULL;
    }

    if (!extend( pool, NumHunks ) ) {
        free( pool->freedHunks );
        free( pool );
        return NULL;
    }

    return pool;
}

void pool_destroy( Pool* pool ) {
    Block* thisBlock;
    Block* nextBlock;

    if (pool == NULL)  return;

    for (thisBlock = pool->block;  thisBlock;   thisBlock = nextBlock) {
        free( thisBlock->base );
        nextBlock = thisBlock->next;
        free( thisBlock );
    }

    free( pool->freedHunks );
    free( pool             );
}

void* pool_getHunk( Pool* pool, bool autoExtend ) {
    void*  ret;
    Block* block;

    if (pool->numFreedHunks > 0) {
        --pool->numFreedHunks;
        ret = pool->freedHunks[ pool->numFreedHunks ];

        pool->numItemsActive++;

        memset( ret, 0, pool->hunkLength );

        return ret;
    }

    if (!(block = pool->curBlock)) {
        return NULL;
    }

    if (block->free < pool->hunkLength) {
        if (!autoExtend) {
            return NULL;
        } else {
            if (!extend( pool, pool->autoExtendNumItems ))   return NULL;
            block = pool->curBlock;
        }
    }

    ret = (void*) block->ptr;

    block->free -= pool->hunkLength;
    block->ptr  += pool->hunkLength;

    memset( ret, 0, pool->hunkLength );

    ++ pool->numItemsActive;

    return ret;
}

static bool freeHunk( Pool* pool, void* hunk ) {

    if ( pool->numFreedHunks >= pool->maxNumFreedHunks ) {
        void** newFreedHunks = malloc( (pool->maxNumFreedHunks << 1) * sizeof(void*) );
        if (!newFreedHunks)   return 0;
        memcpy( newFreedHunks, pool->freedHunks, pool->maxNumFreedHunks * sizeof(void*) );
        free( pool->freedHunks );
        pool->freedHunks = newFreedHunks;
        pool->maxNumFreedHunks <<= 1;
    }

    pool->freedHunks[ pool->numFreedHunks ] = hunk;

    ++pool->numFreedHunks;
    --pool->numItemsActive;

    return 1;
}

void* autoPool_getHunk( Pool** pool,  int* numHunks,   int hunkSize ) {
    void* hunk;

    if (!*pool) {
        *numHunks = 0;
        *pool = pool_create( hunkSize, 4096, 1024 );    assert( *pool );
    }

    hunk = pool_getHunk( *pool, TRUE );

    *numHunks += 1;

    return hunk;
}

void  autoPool_freeHunk( Pool** pool, int* numHunks, void* hunk ) {
    assert( *pool && *numHunks >= 1 );

    freeHunk( *pool, hunk );

    *numHunks -= 1;

    if (*numHunks == 0) {
        pool_destroy( *pool );
        *pool = NULL;
    }
}

void autoPool_destroy( Pool** pool,int* numHunks ) {
    if (*pool) {
        pool_destroy( *pool );
    }
    *pool = NULL;
    *numHunks = 0;
}
