//-{------------------------------------------------------------------------

#include "pzip.h"
#include "arithc.h"
#include <time.h>

#include <stdio.h>

#include "det.h"
#include "contexttrie.h"
#include "context.h"
#include "exclude.h"
#include "minusone.h"
#include "config.h"
#include "codecntx.h"
#include "loe.h"

#define DO_STATS

typedef struct {
    ContextTrie* contextTrie;
    Arith*       arith;
    Exclude*     exclude;

    See*         see;
    Det*         det;
} Pzip;

static Pzip* pzip_create( void ) {
    Pzip*  pzip;
    uint   alphabets[ PZIP_ORDER ];

    pzip = new( Pzip );    assert( pzip );

    alphabets[0] = alphabets[1] = alphabets[2] = alphabets[3] = alphabets[4] = 256;
    alphabets[5] = 256 * 256 * 256;
    alphabets[6] = alphabets[7] = UINT_MAX;      assert( PZIP_ORDER == 8 );

    pzip->contextTrie = contextTrie_create( alphabets, PZIP_ORDER, PZIP_TRIE_MEGS );
    assert( pzip->contextTrie );

    pzip->arith = arith_init();

    pzip->exclude = exclude_create( 256 );

    pzip->see = see_create();

    pzip->det = det_create( PZIP_DET_MEGS );

    return pzip;
}

static void pzip_destroy( Pzip* pzip ) {

    exclude_destroy( pzip->exclude );
    arith_destroy(   pzip->arith   );
    see_destroy(     pzip->see     );

    /* Doing this is much faster, but ruins any encapsulation     */
    /* we might have had : eg. we destroy the context contextTrie */
    /* for ALL pzip objects:                                      */
    contextTrie_destroy( pzip->contextTrie );
    context_destroyAllContexts();

    det_destroy( pzip->det );

    destroy( pzip );
}

uint pzip_encodeArray( ubyte* rawBuf, uint rawLen, ubyte* compBuf ) {
    Pzip*  pzip;
    Arith* arith;
    uint   compLen;
    int    i;
    ubyte* rawPtr;
    ubyte* rawBufEnd;
    clock_t began_at;

    #ifdef DO_STATS
    int numChoseLOE[ PZIP_ORDER+1 ],
        numTriedByOrder[ PZIP_ORDER+1 ],
        numCodedByOrder[ PZIP_ORDER+1 ]
    ;
    int numCodedDet = 0;
    #endif

    pzip = pzip_create();

    rawPtr    = rawBuf;
    rawBufEnd = rawBuf + rawLen;

    began_at = clock();

    /* Seed a preamble: */
    memcpy( compBuf, rawPtr, PZIP_SEED_BYTES );
    assert( PZIP_SEED_BYTES > 0 );
    memset( rawPtr - PZIP_MAX_CONTEXT_LEN, PZIP_SEED_BYTE, PZIP_MAX_CONTEXT_LEN );
    rawPtr  += PZIP_SEED_BYTES;

    arith = pzip->arith;
    arith_encodeInitNoStuff( arith, compBuf + PZIP_SEED_BYTES );

#ifdef DO_STATS
    memset(numChoseLOE,0,(PZIP_ORDER+1)*sizeof(int));
    memset(numTriedByOrder,0,(PZIP_ORDER+1)*sizeof(int));
    memset(numCodedByOrder,0,(PZIP_ORDER+1)*sizeof(int));
#endif

    while (rawPtr < rawBufEnd) {
        ulong cntx;
        ulong tindeces[    PZIP_ORDER    ];
        Context* contexts[ PZIP_ORDER +1 ];
        int sym, LOE_Order, order, codedOrder;
        bool useFull;

        exclude_clear( pzip->exclude );

        sym = *rawPtr;

        /* Build the forward context: */

        cntx = getulong( rawPtr -4 );

        /* The context structure is determined by these indices; */
        /* We do 1-2-3-4-5-8-12-16 , all with the contextTrie.   */ 

        tindeces[0] = rawPtr[-1];       tindeces[1] = rawPtr[-2];
        tindeces[2] = rawPtr[-3];       tindeces[3] = rawPtr[-4];
        tindeces[4] = rawPtr[-5];
        tindeces[5] = rawPtr[-6] + (rawPtr[-7]<<8) + (rawPtr[-8]<<16);
        tindeces[6] = getulong(rawPtr-12);
        tindeces[7] = getulong(rawPtr-16);

        /* Get all the contexts from the contextTrie.                       */
        /* We must do this before the det, cuz the det points out of        */
        /* the top context node (we don't need to do this in "follow" mode) */
        contextTrie_getNodes( pzip->contextTrie, tindeces, contexts );

        /* Note: Including Det in the LOE helps compression on text-like files */
        /*       (paper1,bib) but hurts on LZ-like files (trans,progs)         */
        /*       since it hurts speed so much, we don't do it                  */
        useFull = TRUE;

        if (!det_encode( pzip->det, arith, rawPtr, rawBuf, sym, pzip->exclude, contexts[PZIP_ORDER], &useFull )) {

            /* Do the initial LOE to pick a start order: */
            LOE_Order = loe_chooseOrder( contexts, cntx, PZIP_ORDER, pzip->exclude, pzip->see, useFull );

            #ifdef DO_STATS
            ++ numChoseLOE[ LOE_Order ];
            #endif

            /* Go down the orders: */
            for (order = LOE_Order; ; ) {

                /* Try to coder from order: */

                #ifdef DO_STATS
                ++ numTriedByOrder[ order ];
                #endif

                if (codeContext_encodeFromContext( arith, contexts[order], pzip->exclude, pzip->see, cntx, sym, &useFull )) {
                    #ifdef DO_STATS
                    ++ numCodedByOrder[ order ];
                    #endif
                    goto encoded;
                }
                        
                if (order == 0)   break;
                                        
                /* Maybe skip down a few: */
                order = loe_chooseOrder( contexts, cntx, order-1, pzip->exclude, pzip->see,useFull );
            }

            /* Encode raw with order -1: */
            encodeOrderMinusOne( sym, 256, arith, pzip->exclude );

            /* Did encode, now update the stats: */

        encoded:
                        
            codedOrder = max( order, 0 );

            for (order = 0;   order <= PZIP_ORDER;   order++) {
                context_update( contexts[order], sym, cntx, pzip->see, codedOrder );
            }
        } else {
            #ifdef DO_STATS
            numCodedDet++;
            #endif
        }

        det_update( pzip->det, rawPtr, rawBuf, sym );

        ++ rawPtr;

        i = rawPtr - rawBuf;
        if (i % PZIP_PRINTF_INTERVAL == 0   &&   verbose) {
            fprintf(stderr, "%d/%d\r", i, rawLen );
            fflush( stderr );
        }
    }
    if (verbose) {
        clock_t clocks  = clock() - began_at;                        /* Do NOT combine   */
        double  secs    = (double)clocks / (double)CLOCKS_PER_SEC;   /* these two lines! */
        fprintf(stderr, "%d/%d\n", rawLen, rawLen );
        fprintf(stderr,"%s : %f secs = %2.1f %ss/sec\n", "encode", secs, (double)rawLen / secs, "byte" );
    }

    compLen = arith_encodeDone( arith ) + PZIP_SEED_BYTES;

    pzip_destroy( pzip );

    /* The arithc has stuffed 1 byte; put it back: */ 
    compBuf[ PZIP_SEED_BYTES-1 ] = rawBuf[ PZIP_SEED_BYTES-1 ];

    #ifdef DO_STATS
    if (verbose ) {
        printf( "o : %7s : %7s : %7s\n", "loe", "tried", "coded" );
        printf("d : %7d : %7d : %7d\n", rawLen, rawLen, numCodedDet );
        for (i = PZIP_ORDER;   i >= 0;   i--) {
            printf(
                "%d : %7d : %7d : %7d\n",
                i, numChoseLOE[i], numTriedByOrder[i], numCodedByOrder[i]
            );
        }
    }
    #endif

    return compLen;
}

bool pzip_decodeArray( ubyte* rawBuf, uint rawLen, ubyte* compBuf ) {
    Arith* arith;
    int    i;

    Pzip*  pzip      = pzip_create();
    ubyte* rawPtr    = rawBuf;
    ubyte* rawBufEnd = rawBuf + rawLen;
    clock_t began_at;

    memcpy( rawPtr, compBuf, PZIP_SEED_BYTES );
    memset( rawPtr - PZIP_MAX_CONTEXT_LEN, PZIP_SEED_BYTE, PZIP_MAX_CONTEXT_LEN );

    rawPtr  += PZIP_SEED_BYTES;
    compBuf += PZIP_SEED_BYTES;

    arith = pzip->arith;
    arith_decodeInitNoStuff( arith, compBuf );

    began_at = clock();

    while (rawPtr < rawBufEnd) {
        ulong     cntx;
        ulong     tindeces[ PZIP_ORDER ];
        Context * contexts[ PZIP_ORDER + 1 ];
        int       sym, LOE_Order, order, codedOrder;
        bool      useFull;

        exclude_clear( pzip->exclude );

        /* Build the forward context: */
        cntx = getulong( rawPtr - 4 );

        tindeces[0] = rawPtr[-1];       tindeces[1] = rawPtr[-2];
        tindeces[2] = rawPtr[-3];       tindeces[3] = rawPtr[-4];
        tindeces[4] = rawPtr[-5];
        tindeces[5] = rawPtr[-6] + (rawPtr[-7]<<8) + (rawPtr[-8]<<16);
        tindeces[6] = getulong(rawPtr-12);
        tindeces[7] = getulong(rawPtr-16);

        /* Get all the contexts from the contextTrie: */
        contextTrie_getNodes( pzip->contextTrie, tindeces, contexts );

        useFull = TRUE;

        if (!det_decode( pzip->det, arith, rawPtr, rawBuf, &sym, pzip->exclude, contexts[PZIP_ORDER], &useFull )) {

            /* Do the initial LOE to pick a start order: */
            LOE_Order = loe_chooseOrder( contexts, cntx, PZIP_ORDER, pzip->exclude, pzip->see, useFull );

            /* Go down the orders: */
            for (order = LOE_Order;; ) {

                /* Try to coder from order: */
                if (codeContext_decodeFromContext( arith, contexts[order], pzip->exclude, pzip->see, cntx, &sym, &useFull )) {
                    goto decoded;
                }
                        
                if (order == 0)   break;
                                        
                /* Maybe skip down a few: */
                order = loe_chooseOrder( contexts, cntx, order-1, pzip->exclude, pzip->see,useFull );
            }

            /* Decode raw with order -1: */
            sym = decodeOrderMinusOne( 256, arith, pzip->exclude );

            /* Did decode, now update the stats: */
        decoded:
                        
            codedOrder = max( order, 0 );

            for (order = 0;   order <= PZIP_ORDER;   ++order) {
                context_update( contexts[order], sym, cntx, pzip->see, codedOrder );
            }
        }

        det_update( pzip->det, rawPtr, rawBuf, sym );

        *rawPtr++ = sym;
                
        i = rawPtr - rawBuf;
        if (i % PZIP_PRINTF_INTERVAL == 0   &&   verbose) {
            fprintf(stderr, "%d/%d\r", i, rawLen );
            fflush( stderr );
        }
    }

    if (verbose) {
        clock_t clocks  = clock() - began_at;                        /* Do NOT combine   */
        double  secs    = (double)clocks / (double)CLOCKS_PER_SEC;   /* these two lines! */
        fprintf(stderr, "%d/%d\n", rawLen, rawLen );
        fprintf(stderr,"%s : %f secs = %2.1f %ss/sec\n", "decode", secs, (double)rawLen / secs, "byte" );
    }

    arith_decodeDone( arith );

    pzip_destroy( pzip );

    return TRUE;
}

