/********************************************

 secondary escape estimation (See)

 this version does 3-order weighting by entropy

********************************************/

#include "see.h"
#include "config.h"
#include "context.h"

#include "intmath.h"
#include "pool.h"

//-}{----------- defines ------------------------------------

// the hash is:
// top 5 bits are esc/tot
// next 2 are order
// then 2 for # of chars in parent cntx
// then four pairs of 2 from cntx
// then 5 bits from order0

#define ORDER0_BITS ( 9)        /* <> These cutoffs and the hashes could all be tuned. */
#define ORDER1_BITS (16)
#define ORDER2_BITS (23)

#define ORDER0_SIZE (1 << ORDER0_BITS)
#define ORDER1_SIZE (1 << ORDER1_BITS)
#define ORDER2_SIZE (1 << ORDER2_BITS)

#define MAX_SEE_ESCC    ( 3)
#define MAX_SEE_TOTC    (64)

struct SeeState {
    Node      node;
    SeeState* parent;
    SeeState* child;
    uint      hash;
    uint      seen;
    uint      esc;
    uint      tot;
};

struct See {
    SeeState order0[ ORDER0_SIZE ];
    SeeState order1[ ORDER1_SIZE ];
    Pool*    StatePool;
};

void SeeState_StatsFromHash( SeeState *ss, uint FiveBits );

static void precondition( See* see ) {
    uint escC, totCH;
    uint escapeCount, totSymCount;

    for (escC = 0;   escC <= 3;   ++escC) {
        escapeCount = escC + 1;
        for (totCH = 0;   totCH <= 7;   ++totCH) {
            uint h, h_hi, h_lo;
            uint seedEsc, seedTot, totC;
            uint shift;

            switch (totCH) {
            case 0:
            case 1:
            case 2:    totC = totCH;   break;
            case 3:    totC =  3;      break;
            case 4:    totC =  5;      break;
            case 5:    totC =  8;      break;
            case 6:    totC = 11;      break;
            case 7:    totC = 20;      break;
            }

            totSymCount = totC + escapeCount;

            /* The five bit esc/totC: */
            h = (escC << 3) + totCH;
                        
            h_hi = h;

            seedEsc = escapeCount * SEE_INIT_SCALE + SEE_INIT_ESC;
            seedTot = (escapeCount + totSymCount) * SEE_INIT_SCALE + SEE_INIT_TOT;

            shift = ORDER1_BITS - 5;

            for (h_lo = 0;   h_lo < (1UL << shift);   ++h_lo) {
                SeeState* ss;

                h = (h_hi<<shift) | h_lo;

                assert( h < ORDER1_SIZE );

                ss = &see->order1[h];

                ss->esc = seedEsc;
                ss->tot = seedTot;
                ss->hash = h;
                                
                ss->parent = &see->order0[ (h>>(ORDER1_BITS - ORDER0_BITS)) ];
                ss = ss->parent;
                ss->hash = (h>>(ORDER1_BITS - ORDER0_BITS));
                ss->esc = seedEsc;
                ss->tot = seedTot;
            }
        }
    }
}

See* see_create( void ) {
    See* see = new( See );

    see->StatePool = pool_create( sizeof(SeeState), 4096, 1024 );
        
    precondition( see );
        
    return see;
}

void see_destroy( See* see ) {
    pool_destroy( see->StatePool );
    destroy( see );
}

static void getStats( See* see, SeeState* ss, uint* pEscC, uint* pTotC, uint inEsc, uint inTot ) {
    uint e1,e2,e0,t1,t2,t0,h1,h2,h0,s2,s1,s0;
    uint tot,esc;

    e2 = ss->esc;   t2 = ss->tot;   s2 = ss->seen;   ss = ss->parent; assert(ss);
    e1 = ss->esc;   t1 = ss->tot;   s1 = ss->seen;   ss = ss->parent; assert(ss);
    e0 = ss->esc;   t0 = ss->tot;   s0 = ss->seen;

    /* Spooky: ilog2round (accurate) gives 0.003 bpp over intlog2r (inaccurate for values > 256) */

    h2 = (t2<<12)/(t2 * ilog2round(t2) - e2 * ilog2round(e2) - (t2-e2) * ilog2round(t2-e2) + 1);
    h1 = (t1<<12)/(t1 * ilog2round(t1) - e1 * ilog2round(e1) - (t1-e1) * ilog2round(t1-e1) + 1);
    h0 = (t0<<12)/(t0 * ilog2round(t0) - e0 * ilog2round(e0) - (t0-e0) * ilog2round(t0-e0) + 1);

    /* Give less weight to contexts with only the preconditioned stats. */
    /* This helps a bit; *2,3, or 4 seems the best multiple:            */
    if (s0)   h0 <<= 2;
    if (s1)   h1 <<= 2;
    if (s2)   h2 <<= 2;

#if 0
        tot = (h0 + h1 + h2);
        esc = (e2*h2/t2 + e1*h1/t1 + e0*h0/t0);
#else
        {   uint ei,ti,hi;

            /* Also weight in the esc/tot from the context */
            /*      helps about 0.001 bpc on most files    */

            ei = inEsc; ti = inTot;
            hi = (ti<<12)/(ti * ilog2round(ti) - ei * ilog2round(ei) - (ti-ei) * ilog2round(ti-ei) + 1);

            tot = (h0 + h1 + h2 + hi);
            esc = (e2*h2/t2 + e1*h1/t1 + e0*h0/t0 + ei*hi/ti);
        }
#endif

        while (tot >= 16000) {
            tot >>= 1;
            esc >>= 1;
        }
        if (esc < 1) {
            esc = 1;
        }
        if (tot <= esc) {
            tot = esc + 1;
        }

        *pEscC = esc;
        *pTotC = tot;
}

void  see_encodeEscape( See* see, Arith* arith, SeeState* ss, uint escapeCount, uint totSymCount, bool escape ) {
    if (!ss) {
        arith_encBit( arith, totSymCount, escapeCount + totSymCount, escape );
    } else {
        uint esc,tot;
        /* esc, tot = */ getStats( see, ss, &esc, &tot, escapeCount, escapeCount + totSymCount );
        arith_encBit( arith, esc, tot, !escape );
        see_adjustState( see, ss, escape );
    }
}

bool see_decodeEscape( See* see, Arith* arith, SeeState* ss, uint escapeCount, uint totSymCount) {
    if (!ss) {
        return arith_decBit( arith, totSymCount, escapeCount + totSymCount );
    } else {
        bool escape;
        uint esc,tot;
        /* esc, tot = */ getStats( see, ss, &esc, &tot, escapeCount, escapeCount + totSymCount );
        escape = arith_decBit( arith, esc, tot );
        escape = ! escape;
        see_adjustState( see, ss, escape );
        return escape;
    }
}

uint see_getEscapeP( See* see, SeeState* ss, uint escapeCount, uint totSymCount ) {
    if (!ss) {
        return   (escapeCount << PZIP_INTPROB_SHIFT) / (escapeCount + totSymCount);
    } else {
        uint esc, tot ;/* = */ getStats( see, ss, &esc, &tot, escapeCount, escapeCount + totSymCount );
        return (esc << PZIP_INTPROB_SHIFT) / tot;
    }
}

void see_adjustState( See* see, SeeState* ss, bool escape ) {

    while (ss) {
        ++ ss->seen;

        if (escape) {

            ss->esc += SEE_INC;
            ss->tot += SEE_INC + SEE_ESC_TOT_EXTRA_INC;

        } else {

            /* Forget escapes quickly: */
            if (ss->esc >= SEE_ESC_SCALE_DOWN) {
                ss->esc = (ss->esc >> 1) + 1;
                ss->tot = (ss->tot >> 1) + 2;
            }
            ss->tot += SEE_INC;
        }

        if (ss->tot >= SEE_SCALE_DOWN) {
            ss->esc = (ss->esc >> 1) + 1;
            ss->tot = (ss->tot >> 1) + 2;
            assert( ss->tot < SEE_SCALE_DOWN );
        }

        ss = ss->parent;
    }
}

//-----------------------------------------------------------------------
//      Precondition ; also sets up parent pointers

static void statsFromHash( SeeState* ss, uint FiveBits ) {
    uint escC,totCH;
    uint escapeCount,totSymCount;
    uint seedEsc,seedTot,totC=0; // 2004-04-25-CrT: Initialized to quiet gcc

    escC = FiveBits >> 3;
    totCH = FiveBits & 7;

    switch (totCH) {

    case 0:
    case 1:
    case 2:
        totC = totCH;
        break;

    case 3:
        totC = 3;
        break;

    case 4:
        totC = 5;
        break;

    case 5:
        totC = 8;
        break;

    case 6:
        totC = 11;
        break;

    case 7:
        totC = 20;
        break;
    }

    escapeCount = escC + 1;
    totSymCount = totC + escapeCount;

    seedEsc = escapeCount * SEE_INIT_SCALE + SEE_INIT_ESC;
    seedTot = (escapeCount + totSymCount) * SEE_INIT_SCALE + SEE_INIT_TOT;
        
    ss->esc = seedEsc;
    ss->tot = seedTot;
}

SeeState* see_getState( See* see, uint escapeCount, uint totSymCount, ulong cntx, int order, int numSyms, const Context* context ) {
    uint h, o1;
    uint escC, totC;
    uint z;
    SeeState* ss;
    SeeState* o1ss;

    // Do the hash;
    //      order
    //      escC,
    //      totC,
    //      index
    // use MPS count ?

    escC = escapeCount;
    totC = totSymCount;

    if (totC == 0)   return NULL;

    assert( numSyms >= 1 );
    assert( escC    >= 1 );
    assert( totC >= escC );

    totC -= escC;
    -- escC;

    if (escC > MAX_SEE_ESCC
    || totC >= MAX_SEE_TOTC
    ){
        return NULL;
    }
    
    /***
        
        <> Tune the see hash!
        
        We could make a much fancier hash;  Right now there are lots
        of redundant bits; 
        
        eg. When order = 0 we dont use any bits for index, so we could
                        use more for esc & tot
                conversely when esc is large, we should use fewer bits for order & cntx

        This all makes it much messier to do the precondition...

        ****/

    /* Fill up 15 bits of hash: */
    h = 0;

    /* Two bits for the escC <= 3 */
    assert( escC <= 3 );
    h <<= 2;
    h |= escC;
        
    /* Three bits for totC: */
    h <<= 3;
    if      (totC <=  0)   h |= 0;
    else if (totC <=  1)   h |= 1;
    else if (totC <=  2)   h |= 2;
    else if (totC <=  4)   h |= 3;
    else if (totC <=  6)   h |= 4;
    else if (totC <=  9)   h |= 5;
    else if (totC <= 13)   h |= 6;
    else                   h |= 7;
        
    /* Two bits for the order: */
    h <<= 2;
    if (escC >= 1) {
        if (order >= 3){
            h |= 1;
        }
    } else {
        z = order >> 1;
        if (z > 3) {
            z = 3;
        }
        h |= z;
    }

    /* Two bits for num chars in parent                                        */
    /* This was Malcolm's idea, and it helps *huge* (meaning about 0.02 bpc)   */
    /*      paper2 -> 2.196 and trans -> 1.229 !!!                             */
    /* Maybe I should use the actual-coded-parent by LOE instead of the direct */
    /* parent?  There is a problem there : the LOE decision depends on this!   */
    h <<= 2;
    if (context && context->parent) {
        // @@ use full or upex for numsyms?
        z = context->parent->upex.numSyms;
        if (z > 3) {
            z = 3;
        }
        h |= z;
    }

    /* isdet bool ?                                                       */
    /* Helps a tiny bit (0.001) on files with lots of dets (trans, bib).  */
    /* Doesn't affect others.                                             */
    h <<= 1;
    if (numSyms == 1) {
        h |= 1;
    }

    /* Eight bits from index;  two bits from each the last four bytes: */
    if (order > 0) { h <<= 2; h |= ((cntx>>5)&0x3); }
    if (order > 1) { h <<= 2; h |= ((cntx>>13)&0x3); }
    if (escC <= 1) {
        if (order > 2) { h <<= 2;  h |= ((cntx>>21)&0x3); }
        if (order > 3) { h <<= 2;  h |= ((cntx>>29)&0x3); }
    }

    /* The bottom 5 bits of index[0]: */
    h <<= 5;
    h |= cntx & 31;

    assert( h < ORDER2_SIZE );

    o1 = (h >> (ORDER2_BITS - ORDER1_BITS));
    assert( o1 < ORDER1_SIZE );
    o1ss = &see->order1[o1];
    ss = o1ss->child;
    if (ss) {
        for(;;) {
            if (ss->hash == h) {
                // MTF
                o1ss->child = ss;
                return ss;
            }
            ss = node_next(ss);
            if ( ss == o1ss->child ) {
                break;
            }
        }
    }

    ss = pool_getHunk( see->StatePool, TRUE );
    ss->parent = o1ss;
    ss->hash = h;
    node_null( ss );
    if (o1ss->child) {
        node_addHead(o1ss->child,ss);
    }
    o1ss->child = ss;

    #if 0

    /* Inherit: */
    ss->esc = o1ss->esc;
    ss->tot = o1ss->tot;

    #else

    // A fresh init seems better
    statsFromHash( ss, h >> (ORDER2_BITS - 5) );

    #endif

    return ss;
}


